/**
By Josiah Manson

2x2
|0 1|
|2 3|

3x3
|0 1 2|
|3 4 5|
|6 7 8|

4x4
|0  1   2  3|
|4  5   6  7|
|8  9  10 11|
|12 13 14 15|
**/

#pragma once

#include <math.h>

#include "vect.h"

/// 2x2 matrix
template<class T>
class matrix2
{
public:
	T m[4];

	// functions
	void operator=(const matrix2<T> &a)
	{
		for (int i = 0; i < 4; i++)
			m[i] = a.m[i];
	}

	void setIdentity()
	{
		m[0] = 1;
		m[1] = 0;
		m[2] = 0;
		m[3] = 1;
	}

	static matrix2<T> zero()
	{
		matrix2<T> ret;

		ret.m[0] = 0;
		ret.m[1] = 0;
		ret.m[2] = 0;
		ret.m[3] = 0;

		return ret;
	}

	static matrix2<T> rot(T rad)
	{
		matrix2<T> ret;
		float c = cos(rad);
		float s = sin(rad);

		ret.m[0] = c;
		ret.m[1] = -s;
		ret.m[2] = s;
		ret.m[3] = c;

		return ret;
	}

	matrix2<T> transpose() const
	{
		matrix2<T> ret;

		ret.m[0] = m[0];
		ret.m[1] = m[2];
		ret.m[2] = m[1];
		ret.m[3] = m[3];

		return ret;
	}
	T determinant() const
	{
		const float a11 = m[0], a12 = m[1];
		const float a21 = m[2], a22 = m[3];

		return a11*a22 - a12*a21;
	}
	matrix2<T> inverse() const
	{
		const float det = determinant();
		const float det_inv = 1.0 / det;

		matrix2<T> ret;

		const float a11 = m[0], a12 = m[1];
		const float a21 = m[2], a22 = m[3];

		ret.m[0] = a22;
		ret.m[1] = -a12;
		ret.m[2] = -a21;
		ret.m[3] = a11;

		for (int i = 0; i < 4; i++)
			ret.m[i] *= det_inv;

		return ret;
	}

	matrix2<T> mult(const matrix2<T>& mat) const
	{
		matrix2<T> ret;

		ret.m[0] = m[0] * mat.m[0] + m[1] * mat.m[2];
		ret.m[1] = m[0] * mat.m[1] + m[1] * mat.m[3];
		ret.m[2] = m[2] * mat.m[0] + m[3] * mat.m[2];
		ret.m[3] = m[2] * mat.m[1] + m[3] * mat.m[3];

		return ret;
	}

	vect<2,T> mult(const vect<2,T>& v) const
	{
		vect<2,T> ret;

		ret.v[0] = m[0] * v.v[0] + m[1] * v.v[1];
		ret.v[1] = m[2] * v.v[0] + m[3] * v.v[1];

		return ret;
	}

	matrix2<T> operator*(const matrix2<T>& mat) const {return mult(mat);}
	vect<2,T> operator*(const vect<2,T>& vec) const {return mult(vec);}
	T &operator()(int r, int c) {return m[r*2 + c];}

	void operator+=(const matrix2<T>& mat)
	{
		for (int i = 0; i < 4; i++)
			m[i] += mat.m[i];
	}
	
	matrix2<T> operator+(const matrix2<T>& mat)
	{
		matrix2<T> r;
		for (int i = 0; i < 4; i++)
			r.m[i] = m[i] + mat.m[i];
		return r;
	}
};

typedef matrix2<float> matrix2f;
typedef matrix2<double> matrix2d;

/// 3x3 matrix
template<class T>
class matrix3
{
public:
	T m[9];

	// functions
	void operator=(const matrix3<T> &a)
	{
		for (int i = 0; i < 9; i++)
			m[i] = a.m[i];
	}

	void setIdentity()
	{
		m[0] = 1;
		m[1] = 0;
		m[2] = 0;
		m[3] = 0;
		m[4] = 1;
		m[5] = 0;
		m[6] = 0;
		m[7] = 0;
		m[8] = 1;
	}

	static matrix3<T> zero()
	{
		matrix3<T> ret;

		ret.m[0] = 0;
		ret.m[1] = 0;
		ret.m[2] = 0;
		ret.m[3] = 0;
		ret.m[4] = 0;
		ret.m[5] = 0;
		ret.m[6] = 0;
		ret.m[7] = 0;
		ret.m[8] = 0;

		return ret;
	}

	static matrix3<T> rotaxis(const vect<3,T>& axis, T rad)
	{
		matrix3<T> ret;
		float c = cos(rad);
		float s = sin(rad);
		float t = 1 - c;

		float x = axis.v[0];
		float y = axis.v[1];
		float z = axis.v[2];

		ret.m[0] = t*x*x + c;
		ret.m[1] = t*x*y - s*z;
		ret.m[2] = t*x*z + s*y;
		ret.m[3] = t*x*y + s*z;
		ret.m[4] = t*y*y + c;
		ret.m[5] = t*y*z - s*x;
		ret.m[6] = t*x*z - s*y;
		ret.m[7] = t*y*z + s*x;
		ret.m[8] = t*z*z + c;

		return ret;
	}

	matrix3<T> transpose() const
	{
		matrix3<T> ret;

		ret.m[0] = m[0];
		ret.m[1] = m[3];
		ret.m[2] = m[6];
		ret.m[3] = m[1];
		ret.m[4] = m[4];
		ret.m[5] = m[7];
		ret.m[6] = m[2];
		ret.m[7] = m[5];
		ret.m[8] = m[8];

		return ret;
	}
	T determinant() const
	{
		const float a11 = m[0], a12 = m[1], a13 = m[2];
		const float a21 = m[3], a22 = m[4], a23 = m[5];
		const float a31 = m[6], a32 = m[7], a33 = m[8];

		return a11*a22*a33 + a12*a23*a31 + a13*a21*a32 - a11*a23*a32 - a12*a21*a33 - a13*a22*a31;
	}
	matrix3<T> inverse() const
	{
		const float det = determinant();
		const float det_inv = 1.0 / det;

		matrix3<T> ret;

		const float a11 = m[0], a12 = m[1], a13 = m[2];
		const float a21 = m[3], a22 = m[4], a23 = m[5];
		const float a31 = m[6], a32 = m[7], a33 = m[8];

		ret.m[0] = a33*a22-a32*a23;
		ret.m[1] = -(a33*a12-a32*a13);
		ret.m[2] = a23*a12-a22*a13;
		ret.m[3] = -(a33*a21-a31*a23);
		ret.m[4] = a33*a11-a31*a13;
		ret.m[5] = -(a23*a11-a21*a13);
		ret.m[6] = a32*a21-a31*a22;
		ret.m[7] = -(a32*a11-a31*a12);
		ret.m[8] = a22*a11-a21*a12;

		for (int i = 0; i < 9; i++)
			ret.m[i] *= det_inv;

		return ret;
	}

	matrix3<T> mult(const matrix3<T>& mat) const
	{
		matrix3<T> ret;

		ret.m[0] = m[0] * mat.m[0] + m[1] * mat.m[3] + m[2] * mat.m[6];
		ret.m[1] = m[0] * mat.m[1] + m[1] * mat.m[4] + m[2] * mat.m[7];
		ret.m[2] = m[0] * mat.m[2] + m[1] * mat.m[5] + m[2] * mat.m[8];
		ret.m[3] = m[3] * mat.m[0] + m[4] * mat.m[3] + m[5] * mat.m[6];
		ret.m[4] = m[3] * mat.m[1] + m[4] * mat.m[4] + m[5] * mat.m[7];
		ret.m[5] = m[3] * mat.m[2] + m[4] * mat.m[5] + m[5] * mat.m[8];
		ret.m[6] = m[6] * mat.m[0] + m[7] * mat.m[3] + m[8] * mat.m[6];
		ret.m[7] = m[6] * mat.m[1] + m[7] * mat.m[4] + m[8] * mat.m[7];
		ret.m[8] = m[6] * mat.m[2] + m[7] * mat.m[5] + m[8] * mat.m[8];

		return ret;
	}

	vect<3,T> mult(const vect<3,T>& v) const
	{
		vect<3,T> ret;

		ret.v[0] = m[0] * v.v[0] + m[1] * v.v[1] + m[2] * v.v[2];
		ret.v[1] = m[3] * v.v[0] + m[4] * v.v[1] + m[5] * v.v[2];
		ret.v[2] = m[6] * v.v[0] + m[7] * v.v[1] + m[8] * v.v[2];

		return ret;
	}

	matrix3<T> mult(const T v) const
	{
		matrix3<T> ret;

		for (int i = 0; i < 9; i++)
			ret[i] = m[i] * v;

		return ret;
	}

	matrix3<T> operator*(const matrix3<T>& mat) const {return mult(mat);}
	vect<3,T> operator*(const vect<3,T>& vec) const {return mult(vec);}
	vect<3,T> operator*(const T v) const {return mult(v);}
	T &operator()(int r, int c) {return m[r*3 + c];}

	void operator+=(const matrix3<T>& mat)
	{
		for (int i = 0; i < 9; i++)
			m[i] += mat.m[i];
	}
	
	void operator*=(const T v)
	{
		for (int i = 0; i < 9; i++)
			m[i] *= v;
	}

	matrix3<T> operator+(const matrix3<T>& mat)
	{
		matrix3<T> r;
		for (int i = 0; i < 9; i++)
			r.m[i] = m[i] + mat.m[i];
		return r;
	}

	static matrix3<T> crossProduct(vect<3,T> &a)
	{
		matrix3<T> r;

		r.m[0] = 0;
		r.m[1] = -a[2];
		r.m[2] = a[1];
		r.m[3] = a[2];
		r.m[4] = 0;
		r.m[5] = -a[0];
		r.m[6] = -a[1];
		r.m[7] = a[0];
		r.m[8] = 0;

		return r;
	}
};

typedef matrix3<float> matrix3f;
typedef matrix3<double> matrix3d;


/// 4x4 matrix
template<class T>
class matrix4
{
public:
	T m[16];

	// functions
	void operator=(const matrix4<T> &a)
	{
		for (int i = 0; i < 16; i++)
			m[i] = a.m[i];
	}

	void setIdentity()
	{
		m[0] = 1;
		m[1] = 0;
		m[2] = 0;
		m[3] = 0;

		m[4] = 0;
		m[5] = 1;
		m[6] = 0;
		m[7] = 0;

		m[8] = 0;
		m[9] = 0;
		m[10] = 1;
		m[11] = 0;
		
		m[12] = 0;
		m[13] = 0;
		m[14] = 0;
		m[15] = 1;
	}

	static matrix4<T> zero()
	{
		matrix4<T> ret;

		for (int i = 0; i < 16; i++)
			ret.m[i] = 0;

		return ret;
	}


	matrix4<T> transpose() const
	{
		matrix4<T> ret;

		ret.m[0] = m[0];
		ret.m[1] = m[4];
		ret.m[2] = m[8];
		ret.m[3] = m[12];
		ret.m[4] = m[1];
		ret.m[5] = m[5];
		ret.m[6] = m[9];
		ret.m[7] = m[13];
		ret.m[8] = m[2];
		ret.m[9] = m[6];
		ret.m[10] = m[10];
		ret.m[11] = m[14];
		ret.m[12] = m[3];
		ret.m[13] = m[7];
		ret.m[14] = m[11];
		ret.m[15] = m[15];

		return ret;
	}

	matrix4<T> mult(const matrix4<T>& mat) const
	{
		matrix4<T> ret;

		ret.m[0] = m[0] * mat.m[0] + m[1] * mat.m[4] + m[2] * mat.m[8] + m[3] * mat.m[12];
		ret.m[1] = m[0] * mat.m[1] + m[1] * mat.m[5] + m[2] * mat.m[9] + m[3] * mat.m[13];
		ret.m[2] = m[0] * mat.m[2] + m[1] * mat.m[6] + m[2] * mat.m[10] + m[3] * mat.m[14];
		ret.m[3] = m[0] * mat.m[3] + m[1] * mat.m[7] + m[2] * mat.m[11] + m[3] * mat.m[15]; 
		ret.m[4] = m[4] * mat.m[0] + m[5] * mat.m[4] + m[6] * mat.m[8] + m[7] * mat.m[12];
		ret.m[5] = m[4] * mat.m[1] + m[5] * mat.m[5] + m[6] * mat.m[9] + m[7] * mat.m[13]; 
		ret.m[6] = m[4] * mat.m[2] + m[5] * mat.m[6] + m[6] * mat.m[10] + m[7] * mat.m[14]; 
		ret.m[7] = m[4] * mat.m[3] + m[5] * mat.m[7] + m[6] * mat.m[11] + m[7] * mat.m[15]; 
		ret.m[8] = m[8] * mat.m[0] + m[9] * mat.m[4] + m[10] * mat.m[8] + m[11] * mat.m[12]; 
		ret.m[9] = m[8] * mat.m[1] + m[9] * mat.m[5] + m[10] * mat.m[9] + m[11] * mat.m[13]; 
		ret.m[10] = m[8] * mat.m[2] + m[9] * mat.m[6] + m[10] * mat.m[10] + m[11] * mat.m[14]; 
		ret.m[11] = m[8] * mat.m[3] + m[9] * mat.m[7] + m[10] * mat.m[11] + m[11] * mat.m[15]; 
		ret.m[12] = m[12] * mat.m[0] + m[13] * mat.m[4] + m[14] * mat.m[8] + m[15] * mat.m[12]; 
		ret.m[13] = m[12] * mat.m[1] + m[13] * mat.m[5] + m[14] * mat.m[9] + m[15] * mat.m[13]; 
		ret.m[14] = m[12] * mat.m[2] + m[13] * mat.m[6] + m[14] * mat.m[10] + m[15] * mat.m[14]; 
		ret.m[15] = m[12] * mat.m[3] + m[13] * mat.m[7] + m[14] * mat.m[11] + m[15] * mat.m[15];

		return ret;
	}

	vect<4,T> mult(const vect<4,T>& v) const
	{
		vect<4,T> ret;

		ret.v[0] = m[0] * v.v[0] + m[1] * v.v[1] + m[2] * v.v[2] + m[3] * v.v[3];
		ret.v[1] = m[4] * v.v[0] + m[5] * v.v[1] + m[6] * v.v[2] + m[7] * v.v[3]; 
		ret.v[2] = m[8] * v.v[0] + m[9] * v.v[1] + m[10] * v.v[2] + m[11] * v.v[3]; 
		ret.v[3] = m[12] * v.v[0] + m[13] * v.v[1] + m[14] * v.v[2] + m[15] * v.v[3];

		return ret;
	}

	matrix4<T> mult(const T v) const
	{
		matrix4<T> ret;

		for (int i = 0; i < 16; i++)
			ret[i] = m[i] * v;

		return ret;
	}

	matrix4<T> operator*(const matrix4<T>& mat) const {return mult(mat);}
	vect<4,T> operator*(const vect<4,T>& vec) const {return mult(vec);}
	vect<4,T> operator*(const T v) const {return mult(v);}
	T &operator()(int r, int c) {return m[r*4 + c];}

	void operator+=(const matrix4<T>& mat)
	{
		for (int i = 0; i < 16; i++)
			m[i] += mat.m[i];
	}
	
	void operator*=(const T v)
	{
		for (int i = 0; i < 16; i++)
			m[i] *= v;
	}

	matrix4<T> operator+(const matrix4<T>& mat)
	{
		matrix4<T> r;
		for (int i = 0; i < 16; i++)
			r.m[i] = m[i] + mat.m[i];
		return r;
	}

	void fromRotTrans(matrix3<T> &rot, vect<3,T> &trans)
	{
		m[0] = rot.m[0];
		m[1] = rot.m[1];
		m[2] = rot.m[2];
		m[3] = trans.v[0];

		m[4] = rot.m[3];
		m[5] = rot.m[4];
		m[6] = rot.m[5];
		m[7] = trans.v[1];

		m[8] = rot.m[6];
		m[9] = rot.m[7];
		m[10] = rot.m[8];
		m[11] = trans.v[2];

		m[12] = 0;
		m[13] = 0;
		m[14] = 0;
		m[15] = 1;
	}

	void perspectiveLRBT(double n, double f, double l, double r, double b, double t)
	{
		m[0] = 2*n/(r-l);
		m[1] = 0;
		m[2] = (r+l)/(r-l);
		m[3] = 0;

		m[4] = 0;
		m[5] = 2*n/(t-b);
		m[6] = (t+b)/(t-b);
		m[7] = 0;

		m[8] = 0;
		m[9] = 0;
		m[10] = -(f+n)/(f-n);
		m[11] = -2*f*n/(f-n);

		m[12] = 0;
		m[13] = 0;
		m[14] = -1;
		m[15] = 0;
	}

	void perspectiveRT(double n, double f, double r, double t)
	{
		perspectiveLRBT(n, f, -r, r, -t, t);
	}

	void perspectiveFOV(double n, double f, double fov_deg, double aspect)
	{
		double t = n * tan(fov_deg*.5 * 3.1415926535/180.0);
		perspectiveRT(n, f, t*aspect, t);
	}

	// a function to help implement hand rolled multisampling
	// specify an offset where one pixel offset is given by offx = 1 / width, offy = 1 / height
	void perspectiveFOVoffset(double n, double f, double fov_deg, double aspect, double offx, double offy)
	{
		double t = n * tan(fov_deg*.5 * 3.1415926535/180.0);
		perspectiveLRBT(n, f, (-1 + 2*offx)*t*aspect, (1 + 2*offx)*t*aspect, (-1 + 2*offy)*t, (1 + 2*offy)*t);
	}
};

typedef matrix4<float> matrix4f;
typedef matrix4<double> matrix4d;
